
// QTSnapper1.java
// Andrew Davison, May 2005, ad@fivedots.coe.psu.ac.th

/* The specified quicktime movie is loaded, and the video track 
   identified along with its media informatiion.

   A call to getFrame() extracts a single sample (frame) from 
   the movie into a BufferedImage of type
   BufferedImage.TYPE_3BYTE_BGR, and dimensions 
   FORMAT_SIZE x FORMAT_SIZE. The image has the current time
   in hours:minutes.seconds.milliseconds written on top of it.

   The original dimensions of the image in the movie can be 
   retrieved by calling getImageWidth() and getImageHeight().

   Based on MovieFrameExtractor
   by Chris W. Johnson, February 2002
   http://lists.apple.com/archives/quicktime-java/2002/Feb/msg00062.html

   This version of QTSnapper uses the current time to determine which
   frame is retrieved in getFrame(). This means that frames may be skipped.
   The resulting bug is that the image may contain wrongly coloured
   pixels; this seems to be caused by accessing non-consecutive frames.

   On a test Win98 machine, the reported frame rate is about 16 FPS, 
   although TimeBehavior is asking for 25 FPS. Due to frames being skipped,
   it appears to run faster. The 'apparent FPS is about 31 FPS, and
   the % of skipped frames is close to 50%.
*/

import java.awt.image.*;
import java.awt.*;
import java.io.*;
import java.util.Calendar;
import java.text.SimpleDateFormat;
import java.text.DecimalFormat;

import quicktime.*;
import quicktime.io.*;
import quicktime.qd.*;
import quicktime.std.*;
import quicktime.std.image.*;
import quicktime.std.movies.*;
import quicktime.std.movies.media.*;
import quicktime.util.*;


public class QTSnapper1
{
  private static final int FORMAT_SIZE = 128;
      // size of BufferedImage; should be a power of 2, less than 512,
      // or older graphic cards may get upset when using it as a texture

  private boolean isSessionOpen = false;
  private OpenMovieFile movieFile;
  private Movie movie;
  private Track videoTrack;
  private Media vidMedia;

  private MediaSample mediaSample; // current sample/frame
  private int rowInts;             // no. of ints in each row of a frame
  private int pixData[];           // a pixel array

  // video media information
  private int numSamples;    // number of samples in the movie
  private int timeScale;     // media's time scale
  private int duration;      // duration of the media

  private QDGraphics qdGraphics;  // off-screen drawing surface
  private DSequence dSeq;         // decompressed image sequence

  private int width;         // frame width
  private int height;        // frame height

  private BufferedImage img, formatImg;

  // frame rate variables
  private long startTime;
  private long numFramesMade;
  private DecimalFormat frameDf = new DecimalFormat("0.#");  // 1 dp
  private int prevSampNum;
  private int sampNum = 0;
  private int numCycles = 0;
  private int numSkips = 0;


  public QTSnapper1(String fnm)
  /* Load the quicktime movie from fnm, obtain its video track 
     and media information. Set up the graphics environment, 
     pixel array, and BufferedImage objects for later. 
  */
  {  
    if (!(new File(fnm)).exists()) {
      System.out.println("Cannot locate movie: " + fnm);
      System.exit(0);
    }

    System.out.println("Loading movie: " + fnm + "...");
    try {
      // open a QuickTime session
      QTSession.open();
      isSessionOpen = true;

      // open the movie
      movieFile = OpenMovieFile.asRead( new QTFile(fnm) );
      movie = Movie.fromFile(movieFile);

      // extrack the video track from the movie
      videoTrack = movie.getIndTrackType(1, StdQTConstants.videoMediaType,
                                            StdQTConstants.movieTrackMediaType);
      if (videoTrack == null) {
        System.out.println("Sorry, not a video");
        System.exit(0);
      }

      // get the media data struct. used by the video track
      vidMedia = videoTrack.getMedia();  
      timeScale = vidMedia.getTimeScale();
      duration = vidMedia.getDuration();
      numSamples = vidMedia.getSampleCount();

      // get the first sample in the track
      mediaSample = vidMedia.getSample(0, vidMedia.sampleNumToMediaTime(1).time, 1);

      // store the width and height of the image in the media sample
      ImageDescription imgDesc = (ImageDescription) mediaSample.description;
      width = imgDesc.getWidth();
      height = imgDesc.getHeight();

      // set up a drawing environment for converting future frame images
      qdGraphics = new QDGraphics( imgDesc.getBounds());    
      dSeq = new DSequence(imgDesc, qdGraphics, imgDesc.getBounds(), null, null,
          StdQTConstants.codecFlagUseImageBuffer, 
          StdQTConstants.codecLosslessQuality, CodecComponent.bestFidelityCodec);
          // the sample will be decompressed into qdGraphics

      // set up an array for storing the pixel data
      rowInts = qdGraphics.getPixMap().getRowBytes() >>> 2; 
                   // divide by four to get number of 32-bit ints.
      pixData = new int[rowInts * height];

      // configure the BufferedImage objects
      img = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
      formatImg = new BufferedImage(FORMAT_SIZE, FORMAT_SIZE, 
                                        BufferedImage.TYPE_3BYTE_BGR);
        // use a BufferedImage format that can support dynamic texturing in
        // OpenGL v.1.2 (and above) and D3D 

      // movie.start();

      // initialize the frame rate variables
      startTime = System.currentTimeMillis(); 
      numFramesMade = 0;
    }
    catch (Exception e) {
      System.out.println(e);
      System.exit(0);
    }
  } // end of QTSnapper1()


  synchronized public void stopMovie()
  /* stopMovie() and getFrame() are synchronized so that it's not
     possible to terminate the QuickTime session while a frame is being
     copied from the movie. */
  {
    if (isSessionOpen) {
      System.out.println();

      long totalFrames = (numCycles * numSamples) + sampNum;

      // report percentage of skipped frames
      double skipPerCent = (double)(numSkips * 100) / totalFrames;
      System.out.println("Percentage of frames skipped: " + 
                                frameDf.format(skipPerCent) + "%");  // 1 dp

      // report frame rate details
      // FPS
      long duration = System.currentTimeMillis() - startTime;
      double frameRate = ((double) numFramesMade * 1000.0) / duration;
      System.out.println("FPS: " + frameDf.format(frameRate));  // 1 dp

      // 'apparent' FPS (AFPS)
      double appFrameRate = ((double) totalFrames * 1000.0) / duration;
      System.out.println("AFPS: " + frameDf.format(appFrameRate));  // 1 dp

      QTSession.close();
      isSessionOpen = false;
    }
  } // end of stopMovie()



  synchronized public BufferedImage getFrame()
  /* Return a single sample (a frame) from the movie as a 
     BufferedImage object.

     Unlike getFrame() in QTSnapper, the current time is used to 
     determine which sample/frame to extract. This means that frames
     may be skipped.

     The current time is written on top of the image when it's 
     converted to a BufferedImage.

     The retrieved image contains incorrectly coloured pixels
     when non-consecutive samples are accessed. The 
     same effect can be obtained in QTSnapper by incrementing
     the sample index (sampIdx) by more than 1 on each iteration of 
     getFrame().
  */
  {
    if (!isSessionOpen)
      return null;

    // get the time in seconds since the start of QTSnapper1
    double currTime = ((double)(System.currentTimeMillis() - startTime))/1000.0;

    // use the video's time scale
    int videoCurrTime = ((int)(currTime * timeScale))% duration;

    try {
      // backup the previous sample number
      prevSampNum = sampNum;

      // calculate the sample number
      sampNum = vidMedia.timeToSampleNum(videoCurrTime).sampleNum;
      // System.out.println("" + numCycles + ". Sample num: " + sampNum + 
      //                                     "/" + numSamples);

      if (sampNum == prevSampNum)   // no sample change, so no
        return formatImg;           // need to generate a new image

      if (sampNum < prevSampNum)    // movie has just started over
        numCycles++;

      // record the number of frames skipped
      int skipSize = sampNum - (prevSampNum+1);
      if (skipSize > 0) {  // skipped frame(s)
        // System.out.print("*");
        // System.out.flush();
        numSkips += skipSize;
      }

      // get a single sample starting at the sample number's time
      TimeInfo ti = vidMedia.sampleNumToMediaTime(sampNum);
      mediaSample = vidMedia.getSample(0, ti.time, 1);

      writeToBufferedImage(mediaSample, img);

      // resize img, writing it to formatImg
      Graphics g = formatImg.getGraphics();
      g.drawImage(img, 0, 0, FORMAT_SIZE, FORMAT_SIZE, null); 

      // Overlay curent time on image
      g.setColor(Color.RED);
      g.setFont(new Font("Helvetica", Font.BOLD, 12));
      g.drawString(timeNow(), 5, 14);
      g.dispose();

      numFramesMade++;  // count another frame generated
    }
    catch (Exception e) {
      System.out.println(e);
      formatImg = null;
    }

    return formatImg;
  } // end of getFrame()


  private void writeToBufferedImage(MediaSample mediaSample, BufferedImage img)
                                                       throws Exception
  /* Write the contents of the sample into the BufferedImage

     The data passes through many stages. The basic steps are to
     get a 'raw' image from the current media sample (frame), then
     write it as a pixel array into the DataBuffer part of the
     BufferedImage.
  */
  {
    // extract the raw image from the sample
    RawEncodedImage rawIm = RawEncodedImage.fromQTHandle(mediaSample.data);
    dSeq.decompressFrameS(rawIm, 0);            
         // the raw image is rendered into qdGraphics, being decompressed
         // as it goes
    rawIm.disposeQTObject();

    /* Make the data buffer of the img BufferedImage accessible via
       a pixel array (imgPixelData[]). */
    WritableRaster tile = img.getWritableTile(0, 0);
    int[] imgPixelData = ((DataBufferInt) tile.getDataBuffer()).getData();

    // pull a raw image from qdGraphics (this one will be decompressed)
    RawEncodedImage rawIm2 = qdGraphics.getPixMap().getPixelData(); 
  
    /* Transfer the raw image into the img BufferedImage via 
       a imgPixelData[] pixel array. This is simple if the
       size of the raw image matches the destination pixel array
       (see the else branch), otherwise, only a subset is copied 
       over (see the if branch).
    */
    if (rowInts != width) {
      /* Copy raw image to pixData[], then a subset of
         pixData[] into imgPixelData[]. */
      rawIm2.copyToArray(0, pixData, 0, pixData.length);
      int pixIndex = 0, destPixIndex = 0; 
      for(int row = 0; row < height; row++) {
        int cLimit = pixIndex + width;
        for (int cIndex = pixIndex; cIndex < cLimit;  cIndex++)
          imgPixelData[destPixIndex++] = pixData[cIndex];
        pixIndex += rowInts;
      }
    } 
    else   // copy raw image to imgPixelData[] in one step
      rawIm2.copyToArray(0, imgPixelData, 0, width*height);

    img.releaseWritableTile(0, 0);   // release Raster access to image
  }  // end of writeToBufferedImage()


  private String timeNow()
  // return hours:minutes.seconds.milliseconds
  { 
    SimpleDateFormat sdf = new SimpleDateFormat ("HH:mm.ss.SSS");
    Calendar now = Calendar.getInstance();
    return sdf.format(now.getTime());
  }  // end of timeNow()


  public int getImageWidth()
  {  return width;  }

  public int getImageHeight()
  {  return height;  }

} // end of QTSnapper1 class

