
// QTSnapper2.java
// Andrew Davison, May 2005, ad@fivedots.coe.psu.ac.th

/* The specified quicktime movie is loaded, and the video track 
   identified.

   A call to getFrame() extracts a single sample (frame) from 
   the movie into a BufferedImage of type
   BufferedImage.TYPE_3BYTE_BGR, and dimensions 
   FORMAT_SIZE x FORMAT_SIZE. The image has the current time
   in hours:minutes.seconds.milliseconds written on top of it.

   The next frame is selected by using getNextInterestingTime()
   on the video track. Then a PICT image is 'snapped' from the frame,
   and converted via a lengthy process into a BufferedImage.

   The use of getNextInterestingTime() means that getFrame() reads
   each frame in the movie sequence.

   The original dimensions of the image in the movie can be 
   retrieved by calling getImageWidth() and getImageHeight().

   ----
   Derived from two examples in 
      QuickTime for Java: A Developer's Notebook 
      By Chris Adamson, O'Reilly, Jan 2005
      http://www.oreilly.com/catalog/quicktimejvaadn/

   1. Snapping a PICT from a movie: ConvertToJavaImageBetter.java
   Chapter 5 (Working with QuickDraw) and also
   http://lists.apple.com/archives/quicktime-java/2004/Jul/msg00032.html

   2. Moving through a movie frame-by-frame: BasicQTSnapper.java
   Chapter 2 (Playing Movies)

   ----
   On a test Win98 machine, the reported frame rate is about 9 FPS, 
   although TimeBehavior is asking for 25 FPS. This makes the movie 
   noticably slower compared to QTSnapper and QTSnapper1.
*/


import java.awt.image.*;
import java.awt.*;
import java.io.*;
import java.util.Calendar;
import java.text.SimpleDateFormat;
import java.text.DecimalFormat;

import quicktime.*;
import quicktime.io.*;
import quicktime.qd.*;
import quicktime.std.*;
import quicktime.std.image.*;
import quicktime.std.movies.*;
import quicktime.std.movies.media.*;
import quicktime.util.*;
import quicktime.std.clocks.*;
import quicktime.app.view.*;


public class QTSnapper2
{
  private static final int FORMAT_SIZE = 128;
      // size of BufferedImage; should be a power of 2, less than 512,
      // or older graphic cards may get upset when using it as a texture

  private boolean isSessionOpen = false;
  private OpenMovieFile movieFile;
  private Movie movie;
  private Track videoTrack;
  private GraphicsImporter gi;
  private GraphicsImporterDrawer gid;

  private int width;         // frame width
  private int height;        // frame height

  private BufferedImage formatImg;

  // frame rate variables
  private long startTime;
  private long numFramesMade;
  private DecimalFormat frameDf = new DecimalFormat("0.#");  // 1 dp



  public QTSnapper2(String fnm)
  /* Load the quicktime movie from fnm, obtain its video track and media 
     information. Set up the graphics environment and BufferedImage objects for 
     later. */
  {  
    if (!(new File(fnm)).exists()) {
      System.out.println("Cannot locate movie: " + fnm);
      System.exit(0);
    }

    System.out.println("Loading movie: " + fnm + "...");
    try {
      // open a QuickTime session
      QTSession.open();
      isSessionOpen = true;

      // open the movie
      movieFile = OpenMovieFile.asRead( new QTFile(fnm) );
      movie = Movie.fromFile(movieFile);

      // extrack the video track from the movie
      videoTrack = movie.getIndTrackType(1, StdQTConstants.videoMediaType,
                                            StdQTConstants.movieTrackMediaType);
      if (videoTrack == null) {
        System.out.println("Sorry, not a video");
        System.exit(0);
      }

      // set up graphics importer
      gi = new GraphicsImporter(StdQTConstants.kQTFileTypePicture);
      gid = new GraphicsImporterDrawer(gi);

      // configure the BufferedImage object
      formatImg = new BufferedImage(FORMAT_SIZE, FORMAT_SIZE, 
                                        BufferedImage.TYPE_3BYTE_BGR);
        // use a BufferedImage format that can support dynamic texturing in
        // OpenGL v.1.2 (and above) and D3D  

      // initailize the frame rate variables
      startTime = System.currentTimeMillis(); 
      numFramesMade = 0;
    }
    catch (Exception e) {
      System.out.println(e);
      System.exit(0);
    }
  } // end of QTSnapper2()


  synchronized public void stopMovie()
  /* stopMovie() and getFrame() are synchronized so that it's not
     possible to terminate the QuickTime session while a frame is being
     copied from the movie. */
  {
    if (isSessionOpen) {
      // report frame rate details
      long duration = System.currentTimeMillis() - startTime;
      double frameRate = ((double) numFramesMade * 1000.0) / duration;
      System.out.println("FPS: " + frameDf.format(frameRate));  // 1 dp

      QTSession.close();
      isSessionOpen = false;
    }
  } // end of stopMovie()



  synchronized public BufferedImage getFrame()
  /* The next frame is selected by using getNextInterestingTime()
     on the video track. Then a PICT image is 'snapped' from the frame,
     and converted via a lengthy process into a BufferedImage.
  */
  {
    if (!isSessionOpen)
      return null;

    try {
      // get the time details for the next sample
      TimeInfo ti = videoTrack.getNextInterestingTime(
            StdQTConstants.nextTimeMediaSample, movie.getTime(), 1);

      // move the movie to that sample based on the time
      movie.setTime( new TimeRecord(movie.getTimeScale(), ti.time) );
      // System.out.println("Curr time: " + movie.getTime());

      // take a PICT image
      Pict pict = movie.getPict(movie.getTime());

      Image image = writeToImage(pict);

      // resize image, writing it to formatImg
      Graphics g = formatImg.getGraphics();
      g.drawImage(image, 0, 0, FORMAT_SIZE, FORMAT_SIZE, null); 

      // Overlay curent time on image
      g.setColor(Color.RED);
      g.setFont(new Font("Helvetica", Font.BOLD, 12));
      g.drawString(timeNow(), 5, 14);
      g.dispose();

      numFramesMade++;  // count another frame generated
    }
    catch (Exception e) {
      System.out.println(e);
      formatImg = null;
    }

    return formatImg;
  } // end of getFrame()


  private Image writeToImage(Pict pict) throws Exception
  /* The PICT object is treated as a PICT file and piped
     through the QTImageProducer so an Image can be
     created.
  */
  {
    // add a 512-byte header that the PICT would have if it was a file
    byte[] newPictBytes = new byte[pict.getSize() + 512];
    pict.copyToArray(0, newPictBytes, 512, newPictBytes.length - 512);
    pict = new Pict(newPictBytes);

    // access the PICT via a QT ImageProducer
    DataRef ref = new DataRef(pict, StdQTConstants.kDataRefQTFileTypeTag, "PICT");
    gi.setDataReference(ref);
    QDRect rect = gi.getSourceRect();
    Dimension dim = new Dimension(rect.getWidth(), rect.getHeight());
    QTImageProducer ip = new QTImageProducer(gid, dim);

    // convert ImageProducer data to an Image
    Image image = Toolkit.getDefaultToolkit().createImage(ip);
    width = rect.getWidth();
    height = rect.getHeight();

    return image;
  }  // end of writeToImage()


  private String timeNow()
  // return hours:minutes.seconds.milliseconds
  { 
    SimpleDateFormat sdf = new SimpleDateFormat ("HH:mm.ss.SSS");
    Calendar now = Calendar.getInstance();
    return sdf.format(now.getTime());
  }  // end of timeNow()


  public int getImageWidth()
  {  return width;  }

  public int getImageHeight()
  {  return height;  }

} // end of QTSnapper2 class

